package Sophos::IP::Iface;
use strict;
use warnings;
use Socket;
use Socket6           qw(inet_ntop);
use Net::Interface    qw(:lower :iftype :scope mask2cidr scope type cidr2mask);
use Sophos::IP::Utils qw(mask_addr);
use base 'Exporter';

our @EXPORT;
our @EXPORT_OK = qw(
    interfaces
    is_loopback
    if_index_lookup
);

sub interfaces {
    my %opts = (
       loopback => 0,
       ipv4 => 1,
       ipv6 => 1,
       @_,
    );

    my @all_ifs = Net::Interface->interfaces();

    my (@res, @fams);
    push @fams, AF_INET  if $opts{ipv4};
    push @fams, AF_INET6 if $opts{ipv6}; 

    for my $if (@all_ifs) {
        my $info = $if->info();

        FAMILY: for my $family (@fams) {
            next FAMILY unless exists $info->{$family};

            my $addrs = $info->{$family}{number};
            IP: for (my $i=0; $i < $addrs; $i++) {
                 my $addr = $if->address($family, $i);
                 next IP if !$opts{loopback} && is_loopback($addr);

                 my %extra;
                 if ($family == AF_INET6 && 
                     (scope($addr) == RFC2373_LINKLOCAL())) {

                     $extra{scope} = $if->index;
                 }
                 
                 my $mask = $if->netmask($family, $i);
                 my $cidr = mask2cidr($mask);
                 my $snet = mask_addr($addr, netmask => $mask);

                 push @res, {
                      name         => $if->name,
                      ip           => inet_ntop($family, $addr),
                      netmask      => inet_ntop($family, $mask),
                      subnet       => inet_ntop($family, $snet),
                      network_bits => $cidr,
                      %extra,
                 };
            }
        }
    }
    return @res;
}

sub if_index_lookup {
    my %devices;
    for my $if (Net::Interface->interfaces()) {
        $devices{$if->index} = $if->name;
        $devices{$if->name}  = $if->index;
    }
    return \%devices;
}

sub is_loopback {
    my $addr = shift;
    my $len  = length $addr;

    if ($len == 4) {
        my $byte = unpack("C1", $addr);
        return 1 if $byte == 127;
    }
    else {
        return 1 if type($addr) & IPV6_ADDR_LOOPBACK();
    }
    return 0;
}

1;
