package Sophos::IP::Utils;
use strict;
use warnings;
use Socket;
use Socket6 qw(
    inet_pton inet_ntop getaddrinfo getnameinfo unpack_sockaddr_in6
    pack_sockaddr_in6 NI_NUMERICSERV NI_NAMEREQD
);
use Net::Interface qw(:iftype :scope scope type);
use Carp qw(croak);
use base 'Exporter';

our @EXPORT;
our @EXPORT_OK = qw(
    gethostbyaddr_all
    gethostbyname_all
    inet_addr_sane
    inet_pton_all
    inet_ntop_all
    mask_addr
    network_member
    parse_cidr
    parse_hostport
    parse_ip
);

our %EXPORT_TAGS = (
    parse => [qw(parse_cidr parse_hostport parse_ip)]
);

sub inet_addr_sane {
    my $paddr = shift;

    my $naddr = inet_pton_all($paddr);
    return unless defined $naddr;
    return inet_ntop_all($naddr);
}

sub inet_pton_all {
    my $paddr = shift;
    return unless defined $paddr;

    my $naddr = inet_pton(AF_INET, $paddr);
    return $naddr if defined $naddr;

    $naddr = inet_pton(AF_INET6, $paddr);
    if (defined $naddr) {
        my @i32 = unpack("N4", $naddr);
        if ($i32[0] == 0 && $i32[1] == 0 && $i32[2] == 0xFFFF) {
            $naddr = pack("N", $i32[3]);
        }
    }
    return $naddr;
}

sub inet_ntop_all {
    my $naddr = shift;
    return unless defined $naddr;

    my $addr_len = length $naddr; 
    return inet_ntop(AF_INET, $naddr)  if $addr_len == 4;
    return inet_ntop(AF_INET6, $naddr) if $addr_len == 16;
    return;
}

sub gethostbyaddr_all {
    my $paddr = shift;
    return unless defined $paddr;

    my $naddr = inet_pton_all($paddr);
    return unless defined $naddr;

    my $sa     = _pack_sockaddr(0, $naddr);
    my ($host) = getnameinfo($sa, NI_NUMERICSERV|NI_NAMEREQD);
    return $host;
}

sub gethostbyname_all {
    my $host = shift;
    return unless defined $host;

    my %opts = (
       ipv4 => 1,
       ipv6 => 1,
       @_,
    );

    my $family = AF_UNSPEC;
    unless ($opts{ipv4} and $opts{ipv6}) {
       $family = $opts{ipv6} ? AF_INET6 : AF_INET;
    }

    my @res = getaddrinfo($host, '', $family, SOCK_STREAM, 0);

    my (@recs, $saddr);
    while (@res >= 5) {
       ($family, undef, undef, $saddr, undef, @res) = @res;
       if ($family == AF_INET || $family == AF_INET6) {
             my (undef, $addr) = ($family == AF_INET)
                               ? unpack_sockaddr_in($saddr)
                               : unpack_sockaddr_in6($saddr);
             push @recs, $addr;
       }
    }
    return wantarray ? @recs : $recs[0];
}

#
# Parsing routines
#

sub parse_cidr {
    my ($paddr, %opts) = @_;
    my ($family, $cidr);

    ($paddr, $cidr) = split('/', $paddr, 2);
    return unless defined $cidr && $cidr =~ /^\d+$/;

    ($family, $paddr) = parse_ip($paddr, %opts);
    if (!defined $family || 
        ($family eq 'inet' &&  !($cidr >= 8 && $cidr <= 32)) ||
        ($family eq 'inet6' && !($cidr >= 1 && $cidr <= 128))) {
        return;
    } else {
        my @fields = ($family, "$paddr/$cidr");
        return wantarray ? @fields : "@fields";
    }
}

my $RE_hostname_part = qr/[0-9a-zA-Z][\w-]*/;
my $RE_hostname      = qr/$RE_hostname_part(?:\.$RE_hostname_part)*/;

sub parse_hostport {
    my ($hostport, %opts) = @_;
    my $clean_ip = $opts{no_sbrk} || 0;

    my ($type, $paddr, $scope, $port);

    if ($hostport =~ /^(\d+)$/) {
        $paddr = "127.0.0.1";
        $port = $1;
        $type = 'inet';
    }
    elsif ($hostport =~ /^(.*):(\d+)$/) {
        $paddr = $1;
        $port  = $2;

        ($type, $scope) = _sane_addr($paddr);
        if (!$type && $paddr =~ /^$RE_hostname$/) {
            my @res = getaddrinfo($paddr, $port, AF_UNSPEC, SOCK_STREAM, 0);
            while (@res >= 5) {
               my $family;
               ($family, undef, undef, undef, undef, @res) = @res;
               if ($family == AF_INET || $family == AF_INET6) {
                   if ($family == AF_INET) {
                        $type = 'inet';
                        last;
                   }
                   $type = 'inet6';
               }
            }
            $clean_ip = 1;
        }
    }

    if ($type) {
        my @fields = ($type, _format_addr($paddr, $scope, $clean_ip), $port);
        return wantarray ? @fields : "@fields";
    }

    return;
}

sub parse_ip {
    my ($paddr, %opts) = @_;
    my $clean_ip = $opts{no_sbrk} || 0;

    my ($type, $scope) = _sane_addr($paddr);
    if ($type) {
        my @fields = ($type, _format_addr($paddr, $scope, $clean_ip));
        return wantarray ? @fields : "@fields";
    }

    return;
}

sub mask_addr {
    my ($naddr, %maskopts) = @_;
    croak("Missing address") unless defined $naddr;

    my $addr_len = length $naddr;
    croak("mask_addr: Invalid address")
        unless $addr_len == 4 or $addr_len == 16;

    my $mask;
    if ($maskopts{cidr}) {
        $mask = _cidr2mask($maskopts{cidr}, $addr_len);
    }
    elsif ($maskopts{netmask}) {
        $mask = $maskopts{netmask};
    }

    croak("mask_addr: Missing mask") unless defined $mask;
    return $naddr & $mask;
}

sub network_member {
    my ($paddr, $pmask) = @_;
    return 0 unless defined $pmask && defined $paddr;

    my $naddr = inet_pton_all($paddr);

    my ($network, $mask) = split '/', $pmask, 2;
    return 0 unless defined $mask;

    my $nmask;
    if ($mask =~ /^\d+$/) {
        my $cidrmask = _cidr2mask($mask, length $naddr);
        $nmask = inet_pton_all($network) & $cidrmask;
    }
    else {
        $nmask = inet_pton_all($network) & inet_pton_all($mask);
    }

    return 0 unless defined $nmask;
    return 1 if ($naddr & $nmask) eq $nmask;
    return 0;
}

#
# Internal
#

sub _cidr2mask {
    my ($cidr, $addr_len) = @_;
    croak("cidr2mask: Invalid CIDR") unless $cidr =~ /^\d+$/;

    my $addr_bits = $addr_len * 8;
    croak("cidr2mask: cidr out of range")
        unless $cidr >= 0 and $cidr <= $addr_bits;

    return pack("B*", (1 x $cidr) . (0 x ($addr_bits - $cidr)));
}

sub _format_addr {
    my ($paddr, $scope, $clean) = @_;
    my $addrmask = $clean ? '%s%s' : '[%s%s]';
    return sprintf($addrmask, $paddr, $scope || '');
}

sub _sane_addr {
    my ($paddr) = @_;

    $paddr = $1 if $paddr =~ m{^\[(.+)\]$};

    my $scope_id;
    if ($paddr =~ /:/ && $paddr =~ /^([^%]+)(%.+)$/) {
        $paddr = $1;
        $scope_id = $2;
    }

    # If its a valid ip, perform the zero compression etc
    my $naddr = inet_pton_all($paddr);

    # Could be a hostname
    return unless defined $naddr;

    my $type;
    my $nlen = length $naddr;
    if ($nlen == 16) {
        $type = 'inet6'
            unless $scope_id && (scope($naddr) != RFC2373_LINKLOCAL());
    }
    elsif ($nlen == 4) {
        $type = 'inet'
            unless $scope_id;
    }

    if ($type) {
        # Update the callers address with the sanitized version
        $_[0] = inet_ntop_all($naddr); 
        return ($type, $scope_id);
    }
    return;
}

sub _pack_sockaddr {
    croak("Expected two parameters") unless @_ == 2;
    return length($_[1]) == 4 ? pack_sockaddr_in(@_) : pack_sockaddr_in6(@_);
}

1;
